@@ -21,8 +21,8 @@ Example:
 
 :Job Parameters:
     * **project-type**:
-      Defaults to "freestyle", but "maven" as well as "multijob" or "flow"
-      can also be specified.
+      Defaults to "freestyle", but "maven" as well as "multijob", "flow",
+      "workflow" or "externaljob" can also be specified.
 
     * **defaults**:
       Specifies a set of :ref:`defaults` to use for this job, defaults to
@@ -56,6 +56,10 @@ Example:
       Path for a custom workspace. Defaults to Jenkins default
       configuration.
 
+    * **child-workspace**:
+      Path for a child custom workspace. Defaults to Jenkins default
+      configuration. This parameter is only valid for matrix type jobs.
+
     * **quiet-period**:
       Number of seconds to wait between consecutive runs of this job.
       Defaults to ``0``.
@@ -90,15 +94,25 @@ Example:
       The Logrotate section allows you to automatically remove old build
       history. It adds the ``logrotate`` attribute to the :ref:`Job`
       definition. All logrotate attributes default to "-1" (keep forever).
+      **Deprecated on jenkins >=1.637**: use the ``build-discarder``
+      property instead
+
+    * **raw**:
+      If present, this section should contain a single **xml** entry. This XML
+      will be inserted at the top-level of the :ref:`Job` definition.
 
 """
 
+import logging
 import xml.etree.ElementTree as XML
+
 import jenkins_jobs.modules.base
+from jenkins_jobs.xml_config import remove_ignorable_whitespace
 
 
 class General(jenkins_jobs.modules.base.Base):
     sequence = 10
+    logrotate_warn_issued = False
 
     def gen_xml(self, parser, xml, data):
         jdk = data.get('jdk', None)
@@ -139,6 +153,9 @@ class General(jenkins_jobs.modules.base.Base):
         if 'workspace' in data:
             XML.SubElement(xml, 'customWorkspace').text = \
                 str(data['workspace'])
+        if (xml.tag == 'matrix-project') and ('child-workspace' in data):
+            XML.SubElement(xml, 'childCustomWorkspace').text = \
+                str(data['child-workspace'])
         if 'quiet-period' in data:
             XML.SubElement(xml, 'quietPeriod').text = str(data['quiet-period'])
         node = data.get('node', None)
@@ -152,6 +169,12 @@ class General(jenkins_jobs.modules.base.Base):
                 str(data['retry-count'])
 
         if 'logrotate' in data:
+            if not self.logrotate_warn_issued:
+                logging.warn('logrotate is deprecated on jenkins>=1.637, use '
+                             'the property build-discarder on newer jenkins '
+                             'instead')
+                self.logrotate_warn_issued = True
+
             lr_xml = XML.SubElement(xml, 'logRotator')
             logrotate = data['logrotate']
             lr_days = XML.SubElement(lr_xml, 'daysToKeep')
@@ -162,3 +185,14 @@ class General(jenkins_jobs.modules.base.Base):
             lr_adays.text = str(logrotate.get('artifactDaysToKeep', -1))
             lr_anum = XML.SubElement(lr_xml, 'artifactNumToKeep')
             lr_anum.text = str(logrotate.get('artifactNumToKeep', -1))
+
+        if 'raw' in data:
+            raw(parser, xml, data['raw'])
+
+
+def raw(parser, xml_parent, data):
+    # documented in definition.rst since includes and docs is not working well
+    # For cross cutting method like this
+    root = XML.fromstring(data.get('xml'))
+    remove_ignorable_whitespace(root)
+    xml_parent.append(root)
